#include <windows.h>
#include "version.h"

#define VER_FILEVERSION             AVS_MAJOR_VER,AVS_MINOR_VER,AVS_BUGFIX_VER
#define VER_PRODUCTVERSION          AVS_MAJOR_VER,AVS_MINOR_VER,AVS_BUGFIX_VER
#define VER_COMPANYNAME_STR         "The Public\0"
#define VER_FILEDESCRIPTION_STR     "AviSynth video processing scripting language\0"
#define VER_FILEVERSION_STR         AVS_PPSTR(AVS_MAJOR_VER) ", " AVS_PPSTR(AVS_MINOR_VER) ", " AVS_PPSTR(AVS_BUGFIX_VER) "\0"
#define VER_INTERNALNAME_STR        AVS_FULLVERSION "\0"
#define VER_LEGALCOPYRIGHT_STR      "© 2000-2025 Ben Rudiak-Gould and others\0"
#define VER_ORIGINALFILENAME_STR    "avisynth.dll\0"
#define VER_PRODUCTNAME_STR         AVS_FULLVERSION "\0"
#define VER_PRODUCTVERSION_STR      VER_FILEVERSION_STR
#define VER_COMMENTS_STR            "Homepage: http://avisynth.nl\0"

#ifndef _DEBUG
#define VER_DEBUG                   0
#else
#define VER_DEBUG                   VS_FF_DEBUG
#endif

#define AVS_ICON                    0
#define AVSI_ICON                   1


VS_VERSION_INFO VERSIONINFO
FILEVERSION    	VER_FILEVERSION
PRODUCTVERSION 	VER_PRODUCTVERSION
FILEFLAGSMASK  	VS_FFI_FILEFLAGSMASK
FILEFLAGS      	(VS_FF_PRERELEASE|VER_DEBUG)
FILEOS         	VOS__WINDOWS32
FILETYPE       	VFT_DLL
FILESUBTYPE    	VFT2_UNKNOWN
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904E4"
        BEGIN
            VALUE "CompanyName",      VER_COMPANYNAME_STR
            VALUE "FileDescription",  VER_FILEDESCRIPTION_STR
            VALUE "FileVersion",      VER_FILEVERSION_STR
            VALUE "InternalName",     VER_INTERNALNAME_STR
            VALUE "LegalCopyright",   VER_LEGALCOPYRIGHT_STR
//            VALUE "LegalTrademarks1", VER_LEGALTRADEMARKS1_STR
//            VALUE "LegalTrademarks2", VER_LEGALTRADEMARKS2_STR
            VALUE "OriginalFilename", VER_ORIGINALFILENAME_STR
            VALUE "ProductName",      VER_PRODUCTNAME_STR
            VALUE "ProductVersion",   VER_PRODUCTVERSION_STR
            VALUE "Comments",         VER_COMMENTS_STR
        END
    END

    BLOCK "VarFileInfo"
    BEGIN
        /* The following line should only be modified for localized versions.     */
        /* It consists of any number of WORD,WORD pairs, with each pair           */
        /* describing a language,codepage combination supported by the file.      */
        /*                                                                        */
        /* For example, a file might have values "0x409,1252" indicating that it  */
        /* supports English language (0x409) in the Windows ANSI codepage (1252). */

        VALUE "Translation", 0x409, 1252

    END
END


LANGUAGE LANG_NEUTRAL, SUBLANG_SYS_DEFAULT
AVS_ICON   ICON  "../../distrib/Icons/Ico/AvsDocWindows.ico"
AVSI_ICON  ICON  "../../distrib/Icons/Ico/AvsiDocWindows.ico"
